/*
Module : cnominatim.h
Purpose: Defines the interface and implementation for a set of MFC classes to handle Nominatim queries. Nominatim is the search 
         tool used by OpenStreetMap. For further information on Nominatim, please see http://wiki.openstreetmap.org/wiki/Nominatim.
Created: PJN / 23-10-2010
History: PJN / 07-08-2011 1. Updated the code to support WinHttp for downloads in addition to Wininet. By default the class now
                          uses WinHttp unless you define CNOMINATIM_NOWINHTTP before you include "cnominatim.h"
                          2. Updated copyright details
                          3. Fixed a /analyze compiler warning in the Wininet version of CreateRequest
         PJN / 21-07-2013 1. Updated copyright details
                          2. Updated the code to use the new names of the author's WinHTTP wrapper classes. Thanks to Christopher 
                          Hrabia for reporting this issue.
         PJN / 08-11-2015 1. Updated copyright details.
                          2. Updated code to work with the latest version of the author's WinHTTPWrappers classes.
                          3. Updated code to compile cleanly on VC 2010, 2013 & 2015.
                          4. Reworked the CNominatim::Search and CNominatim::AddressLookup methods to pass the details of the 
                          Nominatim server to use as a MFC CString.
                          5. Added support for street, city, county, state, country, postalcode, countrycodes, bounded,
                          extratags & namedetails parameters to Search method.
                          6. Added support for extratags & namedetails parameters to AddressLookup method. 
                          7. All the classes have been renamed and moved into a Nominatim namespace.
                          8. The AddressLookup method has been renamed to ReverseGeocoding
                          9. Added support for the Nominatim AddressLookup API 
         PJN / 08-10-2017 1. Updated copyright details.
                          2. Replaced NULL throughout the codebase with nullptr. This means that the minimum
                          requirement for the framework is now VC 2010.
                          3. Replaced CString::operator LPC*STR() calls throughout the codebase with
                          CString::GetString calls
                          4. Updated the sample app to include a HTTP user-agent in the request.
                          5. Reworked the classes to optionally compile without MFC. By default the classes
                          now use STL classes and idioms but if you define CNOMINATIM_MFC_EXTENSIONS the classes
                          will revert back to the MFC behaviour.
                          6. Replaced BOOL throughout the codebase with bool
                          7. Added SAL annotations to all the codebase
                          8. Added support for the dedupe flag to the Search method
                          9. Added support for extratags and namedetails to the ReverseGeocoding method
                          10. Eliminated usage of AfxGetAppName from CreateSession method
         PJN / 08-10-2017 1. Renamed CNominatimString typedef to String
                          2. Added typedefs for the CSearchPlace and CLookupPlace arrays
         PJN / 22-11-2018 1. Updated copyright details
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that the code
                          will now only compile on VC 2017 or later.
                          3. Replaced enum throughout the codebase with enum class
                          4. Removed code path which supported CNOMINATIM_MFC_EXTENSIONS define
                          5. Removed code path which supported WINHTTPWRAPPERS_MFC_EXTENSIONS define
                          6. Reworked code to use ATL::CComPtr instead of #import
                          7. Removed code path which supported CNOMINATIM_NOWINHTTP
         PJN / 29-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 16-12-2021 1. Updated copyright details.
                          2. Added support for polygon_threshold in Search query.
                          3. Added support for Polygon output to Reverse query.
                          4. Added support for polygon_threshold in Reverse query.
                          5. Added support for Polygon output to Lookup query.
                          6. Added support for polygon_threshold in Lookup query.
         PJN / 30-12-2021 1. Fixed an issue in _ReverseGeocoding where XML node values were not validated.
         PJN / 05-06-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2010 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Includes ////////////////////////////////////////

#pragma once

#ifndef __CNOMINATIM_H__
#define __CNOMINATIM_H__

#ifndef __msxml6_h__
#pragma message("To avoid this message, please put MsXml6.h in your pre compiled header (normally stdafx.h)")
#include <MsXml6.h>
#endif //#ifndef __msxml6_h__

#ifndef __ATLFILE_H__
#pragma message("To avoid this message, please put atlfile.h in your pre compiled header (normally stdafx.h)")
#include <atlfile.h> 
#endif //#ifndef __ATLFILE_H__

#include "WinHTTPWrappers.h" //If you get a compilation error about this missing header file, then you need to download my WinHttpWrappers classes from http://www.naughter.com/winhttpwrappers.html

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef _SSTREAM_
#pragma message("To avoid this message, please put sstream in your pre compiled header (normally stdafx.h)")
#include <sstream>
#endif //#ifndef _SSTREAM_


//////////////////// Classes //////////////////////////////////////////////////

namespace Nominatim
{
//Typedefs
#ifdef _UNICODE
  using String = std::wstring;
#else
  using String = std::string;
#endif //#ifdef _UNICODE

//Enums
enum class PolygonType
{
  No_Polygon,
  Old_Polygon,
  Geojson_Polygon,
  Kml_Polygon,
  Svg_Polygon,
  Text_Polygon
};

enum class OSMType
{
  Undefined,
  Node,
  Way,
  Relation
};

//The parameters for a Nominatim search query
class CSearchQuery
{
public:
//Member variables
  String m_sUserAgent;                            //The user-agent string to use
  String m_sQueryString;                          //The query string to use
  String m_sAcceptLanguage;                       //If specified the preferred language order for showing search results
  String m_sViewbox;                              //The preferred area to find search results in. Format is "<left>, <top>, <right>, <bottom>"
  PolygonType m_Polygon{PolygonType::No_Polygon}; //How should polygons be returned
  double m_fPolygonThreshold{0};                  //The polygon_threshold value to use
  bool m_bAddressDetails{true};                   //Should we include a breakdown of the address into elements
  String m_sEmail;                                //The email address to include
  String m_sExcludingPlaceIds;                    //A comma separated list of the place_id's you want to skip
  int m_nLimit{-1};                               //Limit the number of results to this limit, -1 implies no limit
  String m_sStreet;                               //Alternative lookup to m_sQueryString
  String m_sCity;                                 //Alternative lookup to m_sQueryString
  String m_sCounty;                               //Alternative lookup to m_sQueryString
  String m_sState;                                //Alternative lookup to m_sQueryString
  String m_sCountry;                              //Alternative lookup to m_sQueryString
  String m_sPostalCode;                           //Alternative lookup to m_sQueryString 
  String m_sCountryCodes;                         //Limit search to a specific country (or a list of countries)
  bool m_bBounded{false};                         //Restrict results to only items contained within the bounding box
  bool m_bExtraTags{false};                       //Should we include addition information in the result if available
  bool m_bNameDetails{false};                     //Should we include a list of alternative names in the results
  bool m_bDedupe{true};                           //Should deduping of results occur
};

//One element of the result of a Nominatim search query
class CSearchPlace
{
public:
//Methods
  void Load(_In_ const ATL::CComPtr<IXMLDOMNode>& placeNode, _In_ PolygonType polygon)
  {
    m_sPlaceID = GetAttributeText(placeNode, L"place_id");
    String sOSMType{GetAttributeText(placeNode, L"osm_type")};
    if (sOSMType == _T("node"))
      m_OSMType = OSMType::Node;
    else if (sOSMType == _T("way"))
      m_OSMType = OSMType::Way;
    else if (sOSMType == _T("relation"))
      m_OSMType = OSMType::Relation;
    else
      m_OSMType = OSMType::Undefined;
    m_sOSMID = GetAttributeText(placeNode, L"osm_id");
    m_sBoundingBox = GetAttributeText(placeNode, L"boundingbox");

    switch (polygon)
    {
      case PolygonType::Old_Polygon:
      {
        m_sPolygon = GetAttributeText(placeNode, L"polygonpoints");
        break;
      }
      case PolygonType::Geojson_Polygon:
      {
        m_sPolygon = GetAttributeText(placeNode, L"geojson");
        break;
      }
      case PolygonType::Kml_Polygon:
      {
        m_sPolygon = GetXMLText(placeNode, L"geokml");
        break;
      }
      case PolygonType::Svg_Polygon:
      {
        m_sPolygon = GetAttributeText(placeNode, L"geosvg");
        break;
      }
      case PolygonType::Text_Polygon:
      {
        m_sPolygon = GetAttributeText(placeNode, L"geotext");
        break;
      }
      default:
      {
        m_sPolygon.clear();
        break;
      }
    }

    String sLatitude{GetAttributeText(placeNode, L"lat")};
    m_fLatitude = _tstof(sLatitude.c_str());
    String sLongitude{GetAttributeText(placeNode, L"lon")};
    m_fLongitude = _tstof(sLongitude.c_str());
    m_sDisplayName = GetAttributeText(placeNode, L"display_name");
    m_sClass = GetAttributeText(placeNode, L"class");
    m_sType = GetAttributeText(placeNode, L"type");
    m_sHouseNumber = GetNodeText(placeNode, L"house_number");
    m_sRoad = GetNodeText(placeNode, L"road");
    m_sVillage = GetNodeText(placeNode, L"village");
    m_sTown = GetNodeText(placeNode, L"town");
    m_sCity = GetNodeText(placeNode, L"city");
    m_sCounty = GetNodeText(placeNode, L"county");
    m_sPostcode = GetNodeText(placeNode, L"postcode");
    m_sCountry = GetNodeText(placeNode, L"country");
    m_sCountryCode = GetNodeText(placeNode, L"country_code");
  }

  static String GetNodeText(_In_ const ATL::CComPtr<IXMLDOMNode>& placeNode, _In_opt_z_ LPCWSTR pszNodeName)
  {
    if (pszNodeName == nullptr)
    {
      ATL::CComBSTR bstrText;
      const HRESULT hr{placeNode->get_text(&bstrText)};
      if (FAILED(hr))
        return String{};
    #ifdef _UNICODE
      return String{bstrText};
    #else
      return String{ATL::CW2A{bstrText}};
    #endif //#ifdef _UNICODE
    }
    else
    {
      ATL::CComPtr<IXMLDOMNode> node;
      HRESULT hr{placeNode->selectSingleNode(ATL::CComBSTR(pszNodeName), &node)};
      if (FAILED(hr))
        return String{};
      if (node)
      {
        ATL::CComBSTR bstrText;
        hr = node->get_text(&bstrText);
        if (FAILED(hr))
          return String{};
      #ifdef _UNICODE
        return String{bstrText};
      #else
        return String{ATL::CW2A{bstrText}};
      #endif //#ifdef _UNICODE
      }
      else
        return String{};
    }
  }

  static String GetXMLText(_In_ const ATL::CComPtr<IXMLDOMNode>& placeNode, _In_opt_z_ LPCWSTR pszNodeName)
  {
    ATL::CComPtr<IXMLDOMNode> node;
    HRESULT hr{placeNode->selectSingleNode(ATL::CComBSTR(pszNodeName), &node)};
    if (FAILED(hr))
      return String{};
    if (node)
    {
      ATL::CComBSTR bstrXML;
      hr = placeNode->get_xml(&bstrXML);
      if (FAILED(hr))
        return String{};
    #ifdef _UNICODE
      return String{bstrXML};
    #else
      return String{ATL::CW2A{bstrXML}};
    #endif //#ifdef _UNICODE
    }
    else
      return String{};
  }

  static String GetAttributeText(_In_ const ATL::CComPtr<IXMLDOMNode>& placeNode, _In_z_ LPCWSTR pszAttributeName)
  {
    ATL::CComPtr<IXMLDOMNamedNodeMap> attributes;
    HRESULT hr{placeNode->get_attributes(&attributes)};
    if (FAILED(hr))
      return String{};

    ATL::CComBSTR bstrAttributeName{pszAttributeName};
    ATL::CComPtr<IXMLDOMNode> attribute;
    hr = attributes->getNamedItem(bstrAttributeName, &attribute);
    if (FAILED(hr))
      return String{};
    if (attribute)
    {
      ATL::CComBSTR bstrText;
      hr = attribute->get_text(&bstrText);
      if (FAILED(hr))
        return String{};
    #ifdef _UNICODE
      return String{bstrText};
    #else
      return String{ATL::CW2A{bstrText}};
    #endif //#ifdef _UNICODE
    }
    else
      return String{};
  }

//Member variables
  String m_sPlaceID;
  OSMType m_OSMType{OSMType::Undefined};
  String m_sOSMID;
  String m_sBoundingBox;
  String m_sPolygon;
  double m_fLatitude{0};
  double m_fLongitude{0};
  String m_sDisplayName;
  String m_sClass;
  String m_sType;
  String m_sHouseNumber;
  String m_sRoad;
  String m_sVillage;
  String m_sTown;
  String m_sCity;
  String m_sCounty;
  String m_sPostcode;
  String m_sCountry;
  String m_sCountryCode;
};

//Typedefs
using CSearchPlaceArray = std::vector<CSearchPlace>;

//The parameters for a Nominatim reverse geocoding query
class CReverseGeocodingQuery
{
public:
//Member variables
  String m_sUserAgent;                            //The user-agent string to use
  String m_sAcceptLanguage;                       //If specified the preferred language order for showing search results
  OSMType m_OSMType{OSMType::Undefined};          //The OSM type of "m_sOSMId"
  String m_sOSMId;                                //The OSM id to lookup
  double m_fLatitude{0};                          //The Latitude to use
  double m_fLongitude{0};                         //The Longitude to use
  int m_nZoom{18};                                //Level of detail required where 0 is country and 18 is house/building
  bool m_bAddressDetails{true};                   //Should we include a breakdown of the address into elements
  String m_sEmail;                                //The email address to include
  bool m_bExtraTags{false};                       //Should we include additional information in the result if available
  bool m_bNameDetails{false};                     //Should we include a list of alternative names in the results
  PolygonType m_Polygon{PolygonType::No_Polygon}; //How should polygons be returned
  double m_fPolygonThreshold{0};                  //The polygon_threshold value to use
};

//The result of a reverse geocoding Nominatim query 
class CReverseGeocodingQueryResult
{
public:
//Methods
  void Load(_In_ const ATL::CComPtr<IXMLDOMNode>& resultNode, _In_ const ATL::CComPtr<IXMLDOMNode>& addressPartsNode)
  {
    m_sPlaceId = CSearchPlace::GetAttributeText(resultNode, L"place_id");
    String sOSMType{CSearchPlace::GetAttributeText(resultNode, L"osm_type")};
    if (sOSMType == _T("node"))
      m_OSMType = OSMType::Node;
    else if (sOSMType == _T("way"))
      m_OSMType = OSMType::Way;
    else if (sOSMType == _T("relation"))
      m_OSMType = OSMType::Relation;
    else
      m_OSMType = OSMType::Undefined;
    m_sOSMId = CSearchPlace::GetAttributeText(resultNode, L"osm_id");
    m_sDisplayName = CSearchPlace::GetNodeText(resultNode, nullptr);
    m_sHouseNumber = CSearchPlace::GetNodeText(addressPartsNode, L"house_number");
    m_sRoad = CSearchPlace::GetNodeText(addressPartsNode, L"road");
    m_sVillage = CSearchPlace::GetNodeText(addressPartsNode, L"village");
    m_sTown = CSearchPlace::GetNodeText(addressPartsNode, L"town");
    m_sCity = CSearchPlace::GetNodeText(addressPartsNode, L"city");
    m_sCounty = CSearchPlace::GetNodeText(addressPartsNode, L"county");
    m_sPostcode = CSearchPlace::GetNodeText(addressPartsNode, L"postcode");
    m_sCountry = CSearchPlace::GetNodeText(addressPartsNode, L"country");
    m_sCountryCode = CSearchPlace::GetNodeText(addressPartsNode, L"country_code");
  }

//Member variables
  String m_sPlaceId;
  OSMType m_OSMType{OSMType::Undefined};
  String m_sOSMId;
  String m_sDisplayName;
  String m_sHouseNumber;
  String m_sRoad;
  String m_sVillage;
  String m_sTown;
  String m_sCity;
  String m_sCounty;
  String m_sPostcode;
  String m_sCountry;
  String m_sCountryCode;
};

//The parameters for a Nominatim Address lookup query
class CAddressLookupQuery
{
public:
//Member variables
  String m_sUserAgent;                            //The user-agent string to use
  String m_sAcceptLanguage;                       //If specified the preferred language order for showing search results
  String m_sOSMIds;                               //The OSM ids to lookup
  bool m_bAddressDetails{true};                   //Should we include a breakdown of the address into elements
  String m_sEmail;                                //The email address to include
  bool m_bExtraTags{false};                       //Should we include additional information in the result if available
  bool m_bNameDetails{false};                     //Should we include a list of alternative names in the results
  PolygonType m_Polygon{PolygonType::No_Polygon}; //How should polygons be returned
  double m_fPolygonThreshold{0};                  //The polygon_threshold value to use
};

//The result of a Nominatim Address lookup query
class CLookupPlace
{
public:
//Methods
  void Load(_In_ const ATL::CComPtr<IXMLDOMNode>& placeNode)
  {
    m_sPlaceId = CSearchPlace::GetAttributeText(placeNode, L"place_id");
    String sOSMType{CSearchPlace::GetAttributeText(placeNode, L"osm_type")};
    if (sOSMType == _T("node"))
      m_OSMType = OSMType::Node;
    else if (sOSMType == _T("way"))
      m_OSMType = OSMType::Way;
    else if (sOSMType == _T("relation"))
      m_OSMType = OSMType::Relation;
    else
      m_OSMType = OSMType::Undefined;
    m_sOSMId = CSearchPlace::GetAttributeText(placeNode, L"osm_id");
    String sLatitude{CSearchPlace::GetAttributeText(placeNode, L"lat")};
    m_fLatitude = _tstof(sLatitude.c_str());
    String sLongitude{CSearchPlace::GetAttributeText(placeNode, L"lon")};
    m_fLongitude = _tstof(sLongitude.c_str());
    m_sDisplayName = CSearchPlace::GetAttributeText(placeNode, L"display_name");
    m_sHouseNumber = CSearchPlace::GetNodeText(placeNode, L"house_number");
    m_sRoad = CSearchPlace::GetNodeText(placeNode, L"road");
    m_sVillage = CSearchPlace::GetNodeText(placeNode, L"village");
    m_sTown = CSearchPlace::GetNodeText(placeNode, L"town");
    m_sCity = CSearchPlace::GetNodeText(placeNode, L"city");
    m_sCounty = CSearchPlace::GetNodeText(placeNode, L"county");
    m_sPostcode = CSearchPlace::GetNodeText(placeNode, L"postcode");
    m_sCountry = CSearchPlace::GetNodeText(placeNode, L"country");
    m_sCountryCode = CSearchPlace::GetNodeText(placeNode, L"country_code");
  }

//Member variables
  String m_sPlaceId;
  OSMType m_OSMType{OSMType::Undefined};
  String m_sOSMId;
  double m_fLatitude{0};
  double m_fLongitude{0};
  String m_sDisplayName;
  String m_sHouseNumber;
  String m_sRoad;
  String m_sVillage;
  String m_sTown;
  String m_sCity;
  String m_sCounty;
  String m_sPostcode;
  String m_sCountry;
  String m_sCountryCode;
};

//More typedefs
using CLookupPlaceArray = std::vector<CLookupPlace>;

//The main class
class CNominatim
{
public:
//Constructors / Destructors
  CNominatim() = default;
  CNominatim(const CNominatim&) = delete;
  CNominatim(CNominatim&&) = delete;
  virtual ~CNominatim()
  {
    ReleaseConnection();
    ReleaseSession();
  }

//Methods
  CNominatim& operator=(const CNominatim&) = delete;
  CNominatim& operator=(CNominatim&&) = delete;

  HRESULT Search(_In_ const CSearchQuery& query, _Inout_ CSearchPlaceArray& result, _In_z_ LPCTSTR pszServer = _T("nominatim.openstreetmap.org"), _In_z_ LPCTSTR pszURL = _T("search")) //Runs the search query
  {
    //Call the helper function to do the main work
    const HRESULT hr{_Search(query, result, pszServer, pszURL)};

    //Free up the connection handles before we return
    ReleaseConnection();

    return hr;
  }

  HRESULT ReverseGeocoding(_In_ const CReverseGeocodingQuery& query, _Inout_ CReverseGeocodingQueryResult& result, _In_z_ LPCTSTR pszServer = _T("nominatim.openstreetmap.org"), _In_z_ LPCTSTR pszURL = _T("reverse")) //Runs the reverse geocoding query
  {
    //Call the helper function to do the main work
    const HRESULT hr{_ReverseGeocoding(query, result, pszServer, pszURL)};

    //Free up the connection handles before we return
    ReleaseConnection();

    return hr;
  }

  HRESULT AddressLookup(_In_ const CAddressLookupQuery& query, _Inout_ CLookupPlaceArray& result, _In_z_ LPCTSTR pszServer = _T("nominatim.openstreetmap.org"), _In_z_ LPCTSTR pszURL = _T("lookup")) //Runs the address lookup query
  {
    //Call the helper function to do the main work
    const HRESULT hr{_AddressLookup(query, result, pszServer, pszURL)};

    //Free up the connection handles before we return
    ReleaseConnection();

    return hr;
  }

  void AttachSession(_In_ HINTERNET hSession) noexcept
  {
    m_Session.Attach(hSession);
  }

  HINTERNET DetachSession() noexcept
  {
    return m_Session.Detach();
  }

protected:
//Methods
  HRESULT _Search(_In_ const CSearchQuery& query, _Inout_ CSearchPlaceArray& result, _In_z_ LPCTSTR pszServer, _In_z_ LPCTSTR pszURL)
  {
    //reset the result array output parameter
    result.clear();

    //Create the internet session
    HRESULT hr{CreateSession(query.m_sUserAgent.c_str())};
    if (FAILED(hr))
      return hr;

    //Create the internet connection
    hr = CreateConnection(pszServer);
    if (FAILED(hr))
      return hr;

    //Next form the request string we will be sending
    String sRequest;
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
    ss << pszURL;
    ss << _T("?format=xml&addressdetails=");
    ss << static_cast<int>(query.m_bAddressDetails);
    ss << _T("&bounded=");
    ss << static_cast<int>(query.m_bBounded);
    ss << _T("&extratags=");
    ss << static_cast<int>(query.m_bExtraTags);
    ss << _T("&namedetails=");
    ss << static_cast<int>(query.m_bNameDetails);
    if (query.m_sQueryString.length())
    {
      ss << _T("&q=");
      ss << query.m_sQueryString;
    }
    else
    {
      if (query.m_sStreet.length())
      {
        ss << _T("&street=");
        ss << query.m_sStreet;
      }
      if (query.m_sCity.length())
      {
        ss << _T("&city=");
        ss << query.m_sCity;
      }
      if (query.m_sCounty.length())
      {
        ss << _T("&county=");
        ss << query.m_sCounty;
      }
      if (query.m_sState.length())
      {
        ss << _T("&state=");
        ss << query.m_sState;
      }
      if (query.m_sCountry.length())
      {
        ss << _T("&country=");
        ss << query.m_sCountry;
      }
      if (query.m_sPostalCode.length())
      {
        ss << _T("&postalcode=");
        ss << query.m_sPostalCode;
      }
    }
    //Add on all the optional parts of the request string
    if (query.m_sAcceptLanguage.length())
    {
      ss << _T("&accept-language=");
      ss << query.m_sAcceptLanguage;
    }
    if (query.m_sViewbox.length())
    {
      ss << _T("&viewbox=");
      ss << query.m_sViewbox;
    }
    if (query.m_sEmail.length())
    {
      ss << _T("&email=");
      ss << query.m_sEmail;
    }
    if (query.m_sExcludingPlaceIds.length())
    {
      ss << _T("&exclude_place_ids=");
      ss << query.m_sExcludingPlaceIds;
    }
    if (query.m_nLimit != -1)
    {
      ss << _T("&limit=");
      ss << query.m_nLimit;
    }
    if (query.m_sCountryCodes.length())
    {
      ss << _T("&countrycodes=");
      ss << query.m_sCountryCodes;
    }
    if (!query.m_bDedupe)
      ss << _T("&dedupe=0");
    switch (query.m_Polygon)
    {
      case PolygonType::Old_Polygon:
      {
        ss << _T("&polygon=1");
        break;
      }
      case PolygonType::Geojson_Polygon:
      {
        ss << _T("&polygon_geojson=1");
        break;
      }
      case PolygonType::Kml_Polygon:
      {
        ss << _T("&polygon_kml=1");
        break;
      }
      case PolygonType::Svg_Polygon:
      {
        ss << _T("&polygon_svg=1");
        break;
      }
      case PolygonType::Text_Polygon:
      {
        ss << _T("&polygon_text=1");
        break;
      }
      default:
      {
        break;
      }
    }
    if (query.m_fPolygonThreshold != 0)
    {
      ss << _T("&polygon_threshold=");
      ss << query.m_fPolygonThreshold;
    }
    sRequest = ss.str();

    //Create and issue the request
    hr = CreateRequest(sRequest.c_str());
    if (FAILED(hr))
      return hr;

    //Read the response
    String sFile;
    hr = ReadResponse(sFile);
    if (FAILED(hr))
      return hr;

    //Load up the XML
    ATL::CComPtr<IXMLDOMDocument2> document;
    hr = LoadResponse(document, sFile.c_str());
    if (FAILED(hr))
      return hr;

    //Pull out all the place nodes from the DOM
    ATL::CComPtr<IXMLDOMNodeList> placeNodes;
    hr = document->selectNodes(ATL::CComBSTR(L"searchresults/place"), &placeNodes);
    if (FAILED(hr))
      return hr;
    if (placeNodes)
    {
      //Iterate across all the places found
      long nPlaces{0};
      hr = placeNodes->get_length(&nPlaces);
      if (FAILED(hr))
        return hr;
      result.reserve(nPlaces);
      for (long i{0}; i<nPlaces; i++)
      {
        //Pull out the current place
        ATL::CComPtr<IXMLDOMNode> placeNode;
        hr = placeNodes->get_item(i, &placeNode);
        if (FAILED(hr))
          return hr;

        //Load it up from the XML
        CSearchPlace place;
        place.Load(placeNode, query.m_Polygon);

        //And put it into the results array
#pragma warning(suppress: 26486 26489)
        result.push_back(place);
      }
    }

    return hr;
  }

  HRESULT _ReverseGeocoding(_In_ const CReverseGeocodingQuery& query, _Inout_ CReverseGeocodingQueryResult& result, _In_z_ LPCTSTR pszServer, _In_z_ LPCTSTR pszURL)
  {
    //Create the internet session
    HRESULT hr{CreateSession(query.m_sUserAgent.c_str())};
    if (FAILED(hr))
      return hr;

    //Create the internet connection
    hr = CreateConnection(pszServer);
    if (FAILED(hr))
      return hr;

    //Next form the request string we will be sending
    String sRequest;
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
    ss << pszURL;
    ss << _T("?format=xml&zoom=");
    ss << query.m_nZoom;
    ss << _T("&addressdetails=");
    ss << static_cast<int>(query.m_bAddressDetails);
    ss << _T("&extratags=");
    ss << static_cast<int>(query.m_bExtraTags);
    ss << _T("&namedetails=");
    ss << static_cast<int>(query.m_bNameDetails);
    //Add on all the optional parts of the request string
    if (query.m_sAcceptLanguage.length())
    {
      ss << _T("&accept-language=");
      ss << query.m_sAcceptLanguage;
    }
    if (query.m_OSMType != OSMType::Undefined)
    {
      switch (query.m_OSMType)
      {
        case OSMType::Node:
        {
          ss << _T("&osm_type=N&osm_id=");
          break;
        }
        case OSMType::Way:
        {
          ss << _T("&osm_type=W&osm_id=");
          break;
        }
        case OSMType::Relation:
        {
          ss << _T("&osm_type=R&osm_id=");
          break;
        }
        default:
        {
#pragma warning(suppress: 26477)
          ATLASSERT(false);
          break;
        }
      }
      ss << query.m_sOSMId;
    }
    else
    {
      ss << _T("&lat=");
      ss << query.m_fLatitude;
      ss << _T("&lon=");
      ss << query.m_fLongitude;
    }
    if (query.m_sEmail.length())
    {
      ss << _T("&email=");
      ss << query.m_sEmail;
    }
    switch (query.m_Polygon)
    {
      case PolygonType::Old_Polygon:
      {
        ss << _T("&polygon=1");
        break;
      }
      case PolygonType::Geojson_Polygon:
      {
        ss << _T("&polygon_geojson=1");
        break;
      }
      case PolygonType::Kml_Polygon:
      {
        ss << _T("&polygon_kml=1");
        break;
      }
      case PolygonType::Svg_Polygon:
      {
        ss << _T("&polygon_svg=1");
        break;
      }
      case PolygonType::Text_Polygon:
      {
        ss << _T("&polygon_text=1");
        break;
      }
      default:
      {
        break;
      }
    }
    if (query.m_fPolygonThreshold != 0)
    {
      ss << _T("&polygon_threshold=");
      ss << query.m_fPolygonThreshold;
    }
    sRequest = ss.str();

    //Create and issue the request
    hr = CreateRequest(sRequest.c_str());
    if (FAILED(hr))
      return hr;

    //Read the response
    String sFile;
    hr = ReadResponse(sFile);
    if (FAILED(hr))
      return hr;

    //Load up the XML
    ATL::CComPtr<IXMLDOMDocument2> document;
    hr = LoadResponse(document, sFile.c_str());
    if (FAILED(hr))
      return hr;

    //Pull out the result from the DOM
    ATL::CComPtr<IXMLDOMNode> resultNode;
    hr = document->selectSingleNode(ATL::CComBSTR(L"reversegeocode/result"), &resultNode);
    if (FAILED(hr))
      return hr;
    if (resultNode == nullptr)
      return E_FAIL;
    ATL::CComPtr<IXMLDOMNode> addressPartsNode;
    hr = document->selectSingleNode(ATL::CComBSTR(L"reversegeocode/addressparts"), &addressPartsNode);
    if (FAILED(hr))
        return hr;
    if (addressPartsNode == nullptr)
      return E_FAIL;
    result.Load(resultNode, addressPartsNode);
    return hr;
  }

  HRESULT _AddressLookup(_In_ const CAddressLookupQuery& query, _Inout_ CLookupPlaceArray& result, _In_z_ LPCTSTR pszServer, _In_z_ LPCTSTR pszURL)
  {
    //reset the result array output parameter
    result.clear();

    //Create the internet session
    HRESULT hr{CreateSession(query.m_sUserAgent.c_str())};
    if (FAILED(hr))
      return hr;

    //Create the internet connection
    hr = CreateConnection(pszServer);
    if (FAILED(hr))
      return hr;

    //Next form the request string we will be sending
    String sRequest;
  #ifdef _UNICODE
    std::wostringstream ss;
  #else
    std::ostringstream ss;
  #endif //#ifdef _UNICODE
    ss << pszURL;
    ss << _T("?osm_ids=");
    ss << query.m_sOSMIds;
    ss << _T("&format=xml&addressdetails=");
    ss << static_cast<int>(query.m_bAddressDetails);
    ss << _T("&extratags=");
    ss << static_cast<int>(query.m_bExtraTags);
    ss << _T("&namedetails=");
    ss << static_cast<int>(query.m_bNameDetails);
    //Add on all the optional parts of the request string
    if (query.m_sAcceptLanguage.length())
    {
      ss << _T("&accept-language=");
      ss << query.m_sAcceptLanguage;
    }
    if (query.m_sEmail.length())
    {
      ss << _T("&email=");
      ss << query.m_sEmail;
    }
    switch (query.m_Polygon)
    {
      case PolygonType::Old_Polygon:
      {
        ss << _T("&polygon=1");
        break;
      }
      case PolygonType::Geojson_Polygon:
      {
        ss << _T("&polygon_geojson=1");
        break;
      }
      case PolygonType::Kml_Polygon:
      {
        ss << _T("&polygon_kml=1");
        break;
      }
      case PolygonType::Svg_Polygon:
      {
        ss << _T("&polygon_svg=1");
        break;
      }
      case PolygonType::Text_Polygon:
      {
        ss << _T("&polygon_text=1");
        break;
      }
      default:
      {
        break;
      }
    }
    if (query.m_fPolygonThreshold != 0)
    {
      ss << _T("&polygon_threshold=");
      ss << query.m_fPolygonThreshold;
    }
    sRequest = ss.str();

    //Create and issue the request
    hr = CreateRequest(sRequest.c_str());
    if (FAILED(hr))
      return hr;

    //Read the response
    String sFile;
    hr = ReadResponse(sFile);
    if (FAILED(hr))
      return hr;

    //Load up the XML
    ATL::CComPtr<IXMLDOMDocument2> document;
    hr = LoadResponse(document, sFile.c_str());
    if (FAILED(hr))
      return hr;

    //Pull out all the place nodes from the DOM
    ATL::CComPtr<IXMLDOMNodeList> placeNodes;
    hr = document->selectNodes(ATL::CComBSTR(L"lookupresults/place"), &placeNodes);
    if (FAILED(hr))
      return hr;
    if (placeNodes)
    {
      //Iterate across all the places found
      long nPlaces{0};
      hr = placeNodes->get_length(&nPlaces);
      if (FAILED(hr))
        return hr;
      result.reserve(nPlaces);
      for (long i{0}; i<nPlaces; i++)
      {
        //Pull out the current place
        ATL::CComPtr<IXMLDOMNode> placeNode;
        hr = placeNodes->get_item(i, &placeNode);
        if (FAILED(hr))
          return hr;

        //Load it up from the XML
        CLookupPlace place;
        place.Load(placeNode);

        //And put it into the results array
#pragma warning(suppress: 26486 26489)
        result.push_back(place);
      }
    }

    return hr;
  }

  virtual HRESULT CreateSession(_In_z_ LPCTSTR pszUserAgent)
  {
    //Create the internet session
    if (m_Session == nullptr)
    {
#pragma warning(suppress: 26477)
      return m_Session.Initialize(ATL::CT2W{pszUserAgent}, WINHTTP_ACCESS_TYPE_DEFAULT_PROXY, WINHTTP_NO_PROXY_NAME, WINHTTP_NO_PROXY_BYPASS, 0);
    }

    return S_OK;
  }

  virtual HRESULT CreateConnection(_In_z_ LPCTSTR pszNominatimServer)
  {
    //Create the internet connection
#pragma warning(suppress: 26477)
    ATLASSERT(m_Connection == nullptr);

    return m_Connection.Initialize(m_Session, ATL::CT2W{pszNominatimServer}, 80);
  }

  virtual HRESULT CreateRequest(_In_z_ LPCTSTR pszRequest)
  {
    //Get the temp path
    ATL::CAtlString sTempPath;
    const DWORD dwChars{GetTempPath(_MAX_PATH, sTempPath.GetBuffer(_MAX_PATH))};
    sTempPath.ReleaseBuffer();
    if (dwChars == 0)
      return ATL::AtlHresultFromLastError();

    //Generate a temp file to store the response into
    ATL::CAtlString sFileName;
    const UINT nChars{GetTempFileName(sTempPath, _T("CNM"), 0, sFileName.GetBuffer(_MAX_PATH))};
    sFileName.ReleaseBuffer();
    if (nChars == 0)
      return ATL::AtlHresultFromLastError();

    //Save the file to download into
    m_File.m_sFileToDownloadInto = ATL::CT2W{sFileName};

    return m_File.Initialize(m_Connection, ATL::CT2W{pszRequest});
  }

  virtual HRESULT ReadResponse(_Inout_ String& sFile)
  {
    //Update the output parameter
  #ifdef _UNICODE
    sFile = m_File.m_sFileToDownloadInto;
  #else
    sFile = ATL::CW2A{m_File.m_sFileToDownloadInto.c_str()};
  #endif //#ifdef _UNICODE

    //Do the download
    const HRESULT hr{m_File.SendRequestSync()};
    m_File.ReleaseResources();
    return hr;
  }

  virtual HRESULT LoadResponse(_Inout_ ATL::CComPtr<IXMLDOMDocument2>& document, _In_z_ LPCTSTR pszFile)
  {
    //First create a DOM document
    HRESULT hr{document.CoCreateInstance(__uuidof(DOMDocument60))};
    if (FAILED(hr))
      return hr;

    //Now load up the XML from disk
    hr = document->put_async(VARIANT_FALSE);
    if (FAILED(hr))
      return hr;
    VARIANT_BOOL vbSuccess{VARIANT_FALSE};
    hr = document->load(ATL::CComVariant(pszFile), &vbSuccess);
    if (FAILED(hr))
      return hr;

    //Delete the temp file now that we are finished with it
    DeleteFile(pszFile);

    //Check we could load the xml response ok
    if (vbSuccess == VARIANT_FALSE)
      return ATL::AtlHresultFromWin32(ERROR_FILE_CORRUPT);

    return S_OK;
  }

  virtual void ReleaseConnection() noexcept
  {
    m_File.Close();
    m_Connection.Close();
  }

  virtual void ReleaseSession() noexcept
  {
    m_Session.Close();
  }

  //Member variables
  WinHTTPWrappers::CSession m_Session;
  WinHTTPWrappers::CConnection m_Connection;
  WinHTTPWrappers::CSyncDownloader m_File;
};


}; //namespace Nominatim

#endif //__CNOMINATIM_H__
